import inspect
import pathlib
import sys
import textwrap
from collections import UserList


def is_cyfunction(obj):
    return type(obj).__name__ == "cython_function_or_method"


def is_function(obj):
    return (
        inspect.isbuiltin(obj)
        or is_cyfunction(obj)
        or type(obj) is type(ord)
    )  # fmt: skip


def is_method(obj):
    return (
        inspect.ismethoddescriptor(obj)
        or inspect.ismethod(obj)
        or is_cyfunction(obj)
        or type(obj) in (
            type(str.index),
            type(str.__add__),
            type(str.__new__),
        )
    )  # fmt: skip


def is_classmethod(obj):
    return (
        inspect.isbuiltin(obj)
        or type(obj).__name__ in (
            "classmethod",
            "classmethod_descriptor",
        )
    )  # fmt: skip


def is_staticmethod(obj):
    return type(obj).__name__ == "staticmethod"


def is_datadescr(obj):
    return inspect.isdatadescriptor(obj) and not hasattr(obj, "fget")


def is_property(obj):
    return inspect.isdatadescriptor(obj) and hasattr(obj, "fget")


def is_class(obj):
    return inspect.isclass(obj) or type(obj) is type(int)


class Lines(UserList):
    INDENT = " " * 4
    level = 0

    @property
    def add(self):
        return self

    @add.setter
    def add(self, lines):
        if lines is None:
            return
        if isinstance(lines, str):
            lines = textwrap.dedent(lines).strip().split("\n")
        indent = self.INDENT * self.level
        for line in lines:
            self.append(indent + line)


def signature(obj):
    doc = obj.__doc__
    sig = doc.partition("\n")[0]
    return sig or None


def docstring(obj):
    doc = obj.__doc__
    doc = doc.partition("\n")[2]
    doc = textwrap.dedent(doc).strip()
    doc = f'"""{doc}\n"""'
    doc = textwrap.indent(doc, Lines.INDENT)
    return doc


def visit_constant(constant):
    name, value = constant
    typename = type(value).__name__
    kind = "Constant" if isinstance(value, (str, int, float)) else "Object"
    init = f"_def({typename}, '{name}')"
    doc = f"#: {kind} ``{name}`` of type :class:`{typename}`"
    return f"{name}: {typename} = {init}  {doc}\n"


def visit_function(function):
    sig = signature(function)
    doc = docstring(function)
    body = Lines.INDENT + "..."
    return f"def {sig}:\n{doc}\n{body}\n"


def visit_method(method):
    sig = signature(method)
    doc = docstring(method)
    body = Lines.INDENT + "..."
    return f"def {sig}:\n{doc}\n{body}\n"


def visit_datadescr(datadescr, name=None):
    sig = signature(datadescr)
    doc = docstring(datadescr)
    name = sig.split(":")[0].strip()
    type = sig.split(":")[1].strip()  # noqa: A001
    sig = f"{name}(self) -> {type}"
    body = Lines.INDENT + "..."
    return f"@property\ndef {sig}:\n{doc}\n{body}\n"


def visit_property(prop, name=None):
    gname = prop.fget.__name__
    pname = name or gname
    if pname == gname:
        meth = visit_method(prop.fget)
    else:
        sig = signature(prop.fget)
        name = name or prop.fget.__name__
        type = sig.rsplit("->", 1)[-1].strip()  # noqa: A001
        sig = f"{name}(self) -> {type}"
        doc = docstring(prop)
        body = Lines.INDENT + "..."
        meth = f"def {sig}:\n{doc}\n{body}\n"
    return f"@property\n{meth}"


def visit_constructor(cls, name="__init__", args=None):
    init = name == "__init__"
    argname = cls.__mro__[-2].__name__.lower()
    argtype = cls.__name__
    initarg = args or f"{argname}: {argtype} | None = None"
    selfarg = "self" if init else "cls"
    rettype = "None" if init else "Self"
    arglist = f"{selfarg}, {initarg}"
    sig = f"{name}({arglist}) -> {rettype}"
    ret = "..." if init else "return super().__new__(cls)"
    body = Lines.INDENT + ret
    return f"def {sig}:\n{body}"


def visit_class(cls, done=None):
    skip = {
        "__doc__",
        "__dict__",
        "__module__",
        "__weakref__",
        "__pyx_vtable__",
        "__lt__",
        "__le__",
        "__ge__",
        "__gt__",
        "__str__",
        "__repr__",
    }
    special = {
        "__len__": ("self", "int", None),
        "__bool__": ("self", "bool", None),
        "__hash__": ("self", "int", None),
        "__int__": ("self", "int", None),
        "__index__": ("self", "int", None),
        "__eq__": ("self", "other: object", "/", "bool", None),
        "__ne__": ("self", "other: object", "/", "bool", None),
        "__buffer__": ("self", "flags: int", "/", "memoryview", (3, 12)),
    }
    constructor = (
        "__new__",
        "__init__",
    )

    override = OVERRIDE.get(cls.__name__, {})
    done = set() if done is None else done
    lines = Lines()

    if cls.__name__ == "Exception":
        skip = skip - {f"__{a}{b}__" for a in "lg" for b in "et"}
        constructor = ()
    if "__hash__" in cls.__dict__:
        if cls.__hash__ is None:
            done.add("__hash__")

    base = cls.__base__
    if base is object:
        lines.add = f"class {cls.__name__}:"
    else:
        lines.add = f"class {cls.__name__}({base.__name__}):"
    lines.level += 1

    lines.add = docstring(cls)

    name = "__init__"
    if name not in override:
        sig = signature(cls)
        if sig is not None:
            done.update(constructor)
            args = sig.strip().split("->", 1)[0].strip()
            pidx = args.index("(")
            args = args[pidx:][1:-1]
            lines.add = visit_constructor(cls, name, args)

    for name in constructor:
        if name in done:
            break
        if name in override:
            done.update(constructor)
            lines.add = override[name]
            break
        if name in cls.__dict__:
            done.update(constructor)
            lines.add = visit_constructor(cls, name)
            break

    if "__hash__" in cls.__dict__:
        if cls.__hash__ is None:
            done.add("__hash__")

    dct = cls.__dict__
    keys = list(dct.keys())
    for name in keys:
        if name in done:
            continue

        if name in override:
            done.add(name)
            lines.add = override[name]
            continue

        if name in skip:
            continue

        if name in special:
            done.add(name)
            *args, retv, py = special[name]
            sig = f"{name}({', '.join(args)}) -> {retv}"
            if py is not None:
                lines.add = f"if sys.version_info >= {py}:"
                lines.level += 1
            lines.add = f"def {sig}: ..."
            if py is not None:
                lines.level -= 1
            continue

        attr = getattr(cls, name)

        if is_method(attr):
            done.add(name)
            if name == attr.__name__:
                obj = dct[name]
                if is_classmethod(obj):
                    lines.add = "@classmethod"
                elif is_staticmethod(obj):
                    lines.add = "@staticmethod"
                lines.add = visit_method(attr)
            elif False:
                lines.add = f"{name} = {attr.__name__}"
            continue

        if is_datadescr(attr):
            done.add(name)
            lines.add = visit_datadescr(attr)
            continue

        if is_property(attr):
            done.add(name)
            lines.add = visit_property(attr, name)
            continue

    leftovers = [
        name for name in keys if name not in done and name not in skip
    ]
    if leftovers:
        raise RuntimeError(f"leftovers: {leftovers}")

    lines.level -= 1
    return lines


def visit_module(module, done=None):
    skip = {
        "__doc__",
        "__name__",
        "__loader__",
        "__spec__",
        "__file__",
        "__package__",
        "__builtins__",
    }

    done = set() if done is None else done
    lines = Lines()

    keys = list(module.__dict__.keys())
    keys.sort(key=lambda name: name.startswith("_"))

    constants = [
        (name, getattr(module, name))
        for name in keys
        if all((
            name not in done and name not in skip,
            isinstance(getattr(module, name), int),
        ))
    ]
    for _, value in constants:
        cls = type(value)
        name = cls.__name__
        if name in done or name in skip:
            continue
        if cls.__module__ == module.__name__:
            done.add(name)
            lines.add = visit_class(cls)
            lines.add = ""
    for name, value in constants:
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_constant((name, value))
    if constants:
        lines.add = ""

    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)

        if is_class(value):
            done.add(name)
            lines.add = visit_class(value)
            lines.add = ""
            aliases = [
                (k, getattr(module, k))
                for k in keys
                if all((
                    k not in done and k not in skip,
                    getattr(module, k) is value,
                ))
            ]
            for aliasname, target in aliases:
                done.add(aliasname)
                lines.add = f"{aliasname} = {target.__name__}"
            if aliases:
                lines.add = ""
            instances = [
                (k, getattr(module, k))
                for k in keys
                if all((
                    k not in done and k not in skip,
                    type(getattr(module, k)) is value,
                ))
            ]
            for attrname, attrvalue in instances:
                done.add(attrname)
                lines.add = visit_constant((attrname, attrvalue))
            if instances:
                lines.add = ""
            continue

        if is_function(value):
            done.add(name)
            if name == value.__name__:
                lines.add = visit_function(value)
            else:
                lines.add = f"{name} = {value.__name__}"
            continue

    lines.add = ""
    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_constant((name, value))

    leftovers = [
        name for name in keys if name not in done and name not in skip
    ]
    if leftovers:
        raise RuntimeError(f"leftovers: {leftovers}")
    return lines


IMPORTS = """
import sys
from typing import (
    Any,
    AnyStr,
    Final,
    Literal,
    NoReturn,
)
from typing import (
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Sequence,
    Mapping,
)
try:
    from typing import Self
except ImportError:
    Self = 'Self'
from os import PathLike
"""

HELPERS = """
class _Int(int): pass

def _repr(obj):
    try:
        return obj._name
    except AttributeError:
        return super(obj).__repr__()

def _def(cls, name):
    if cls is int:
       cls = _Int
    obj = cls()
    if cls.__name__ in ('Pickle', 'buffer'):
        return obj
    obj._name = name
    if '__repr__' not in cls.__dict__:
        cls.__repr__ = _repr
    return obj
"""

OVERRIDE = {
    "Exception": {
        "__init__": (
            "def __init__(self, ierr: int = SUCCESS, /) -> None:\n"
            "    return super().__init__(ierr)"
        ),
    },
    "Info": {
        "__iter__": "def __iter__(self) -> Iterator[str]: ...",
        "__getitem__": "def __getitem__(self, item: str, /) -> str: ...",
        "__setitem__": (
            "def __setitem__(self, item: str, value: str, /) -> None: ..."
        ),
        "__delitem__": "def __delitem__(self, item: str, /) -> None: ...",
        "__contains__": "def __contains__(self, value: str, /) -> bool: ...",
    },
    "Op": {
        "__call__": "def __call__(self, x: Any, y: Any, /) -> Any: ...",
    },
    "buffer": {
        "__new__": (
            "def __new__(cls, buf: Buffer, /) -> Self:\n"
            "    return super().__new__(cls)"
        ),
        "__getitem__": (
            "def __getitem__(self, item: int | slice, /) -> int | buffer: ..."
        ),
        "__setitem__": (
            "def __setitem__(self, "
            "item: int | slice, "
            "value:int | Buffer, /) "
            "-> None: ..."
        ),
        "__delitem__": None,
    },
    "__pyx_capi__": None,
    "_typedict": "_typedict: Dict[str, Datatype] = {}",
    "_typedict_c": "_typedict_c: Dict[str, Datatype] = {}",
    "_typedict_f": "_typedict_f: Dict[str, Datatype] = {}",
    "_keyval_registry": None,
}
OVERRIDE.update({
    subtype: {
        "__new__": (
            f"def __new__(cls) -> Self:\n"
            f"    return super().__new__({subtype})"
        ),
        "__repr__": "def __repr__(self) -> str: return self._name",
    }
    for subtype in (
        "BottomType",
        "InPlaceType",
        "BufferAutomaticType",
    )
})  # fmt: skip

TYPING = """
from .typing import *
"""


def visit_mpi4py_MPI():
    from mpi4py import MPI as module

    lines = Lines()
    lines.add = f'"""{module.__doc__}"""'
    lines.add = "# flake8: noqa"
    lines.add = IMPORTS
    lines.add = ""
    lines.add = HELPERS
    lines.add = ""
    lines.add = visit_module(module)
    lines.add = ""
    lines.add = TYPING
    return lines


def generate(filename):
    filename = pathlib.Path(filename)
    filename.parent.mkdir(parents=True, exist_ok=True)
    with filename.open("w", encoding="utf-8") as f:
        for line in visit_mpi4py_MPI():
            print(line, file=f)


def load_module(filename, name=None):
    filename = pathlib.Path(filename)
    if name is None:
        name = filename.stem
    module = type(sys)(name)
    module.__file__ = str(filename)
    module.__package__ = name.rsplit(".", 1)[0]
    preamble = "from __future__ import annotations"
    codelines = filename.read_text(encoding="utf-8").split("\n")
    for lineno, line in enumerate(codelines):
        match = line.strip().startswith
        if not any(map(match, "'\"#")):
            codelines.insert(lineno, preamble)
            break
    code = "\n".join(codelines)
    exec(code, module.__dict__)
    return module


_sys_modules = {}


def replace_module(module):
    name = module.__name__
    assert name not in _sys_modules
    _sys_modules[name] = sys.modules[name]
    sys.modules[name] = module
    pkgname, _, modname = name.rpartition(".")
    if pkgname:
        setattr(sys.modules[pkgname], modname, module)


def restore_module(module):
    name = module.__name__
    assert name in _sys_modules
    sys.modules[name] = _sys_modules[name]
    pkgname, _, modname = name.rpartition(".")
    if pkgname:
        setattr(sys.modules[pkgname], modname, sys.modules[name])


def annotate(dest, source):
    try:
        dest.__annotations__ = source.__annotations__
    except AttributeError:
        pass
    if isinstance(dest, type):
        for name in dest.__dict__.keys():
            if hasattr(source, name):
                obj = getattr(dest, name)
                annotate(obj, getattr(source, name))
    if isinstance(dest, type(sys)):
        for name in dir(dest):
            if hasattr(source, name):
                obj = getattr(dest, name)
                mod = getattr(obj, "__module__", None)
                if dest.__name__ == mod:
                    annotate(obj, getattr(source, name))
        for name in dir(source):
            if not hasattr(dest, name):
                setattr(dest, name, getattr(source, name))


OUTDIR = pathlib.Path("reference")

if __name__ == "__main__":
    generate(OUTDIR / "mpi4py.MPI.py")



import struct

import mpiunittest as unittest

from mpi4py import MPI


@unittest.skipMPI("msmpi")
@unittest.skipMPI("mvapich")
@unittest.skipMPI("impi")
class BaseTestULFM:
    #
    COMM = MPI.COMM_NULL

    def setUp(self):
        self.COMM = self.COMM.Dup()
        self.COMM.Set_errhandler(MPI.ERRORS_RETURN)

    def tearDown(self):
        self.COMM.Free()
        del self.COMM

    def testIsRevoked(self):
        comm = self.COMM
        try:
            self.assertFalse(comm.Is_revoked())
        except NotImplementedError:
            pass

    def testRevoke(self):
        comm = self.COMM
        comm.Is_intra()
        try:
            comm.Revoke()
        except NotImplementedError:
            self.skipTest("mpi-comm_revoke")
        try:
            self.assertTrue(comm.Is_revoked())
        except NotImplementedError:
            pass
        #
        try:
            comm.Barrier()
        except MPI.Exception as exc:
            code = exc.Get_error_class()
        else:
            code = MPI.SUCCESS
        self.assertEqual(code, MPI.ERR_REVOKED)
        #
        try:
            comm.Send([None, 0, MPI.BYTE], MPI.PROC_NULL)
        except MPI.Exception as exc:
            code = exc.Get_error_class()
        else:
            code = MPI.SUCCESS
        self.assertEqual(code, MPI.ERR_REVOKED)
        #
        try:
            comm.Recv([None, 0, MPI.BYTE], MPI.PROC_NULL)
        except MPI.Exception as exc:
            code = exc.Get_error_class()
        else:
            code = MPI.SUCCESS
        self.assertEqual(code, MPI.ERR_REVOKED)

    def testGetFailed(self):
        comm = self.COMM
        group = comm.Get_failed()
        gcmp = MPI.Group.Compare(group, MPI.GROUP_EMPTY)
        group.Free()
        self.assertIn(gcmp, [MPI.IDENT, MPI.CONGRUENT])

    def testAckFailed(self):
        comm = self.COMM
        size = comm.Get_size()
        num_acked = comm.Ack_failed(0)
        self.assertEqual(num_acked, 0)
        num_acked = comm.Ack_failed(size)
        self.assertEqual(num_acked, 0)
        num_acked = comm.Ack_failed()
        self.assertEqual(num_acked, 0)

    def testAgree(self):
        comm = self.COMM
        for i in range(5):
            flag = i
            flag = comm.Agree(flag)
            self.assertEqual(flag, i)

    def testIAgree(self):
        comm = self.COMM
        with self.assertRaises(TypeError):
            comm.Iagree(0)
        with self.assertRaises(ValueError):
            comm.Iagree(bytearray(8))
        ibuf = MPI.buffer.allocate(struct.calcsize("i"))
        flag = memoryview(ibuf).cast("i")
        for i in range(5):
            flag[0] = i
            request = comm.Iagree(flag)
            request.Wait()
            self.assertEqual(flag[0], i)
        size = comm.Get_size()
        if comm.Is_intra() and size > 1:
            for root in range(size):
                rank = comm.Get_rank()
                comm.Barrier()
                if rank == root:
                    ival = int("1011", base=2)
                    flag[0] = ival
                    request = comm.Iagree(flag)
                    self.assertFalse(request.Test())
                    self.assertEqual(flag[0], ival)
                    comm.Barrier()
                else:
                    ival = int("1101", base=2)
                    flag[0] = ival
                    comm.Barrier()
                    self.assertEqual(flag[0], ival)
                    request = comm.Iagree(flag)
                request.Wait()
                ival = int("1001", base=2)
                self.assertEqual(flag[0], ival)

    def testShrink(self):
        comm = self.COMM
        shrink = comm.Shrink()
        self.assertEqual(comm.Get_size(), shrink.Get_size())
        self.assertEqual(comm.Get_rank(), shrink.Get_rank())
        if shrink.Is_inter():
            self.assertEqual(comm.Get_remote_size(), shrink.Get_remote_size())
        shrink.Free()

    def testIShrink(self):
        comm = self.COMM
        shrink, request = comm.Ishrink()
        self.assertTrue(request)
        request.Wait()
        self.assertFalse(request)
        self.assertEqual(comm.Get_size(), shrink.Get_size())
        self.assertEqual(comm.Get_rank(), shrink.Get_rank())
        if shrink.Is_inter():
            self.assertEqual(comm.Get_remote_size(), shrink.Get_remote_size())
        shrink.Free()


class TestULFMSelf(BaseTestULFM, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestULFMWorld(BaseTestULFM, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


@unittest.skipMPI("openmpi(>=5.0.0,<5.0.4)")
@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, "mpi-world-size<2")
class TestULFMInter(BaseTestULFM, unittest.TestCase):
    #
    @classmethod
    def setUpClass(cls):
        BASECOMM = MPI.COMM_WORLD
        size = BASECOMM.Get_size()
        rank = BASECOMM.Get_rank()
        if rank < size // 2:
            COLOR = 0
            LOCAL_LEADER = 0
            REMOTE_LEADER = size // 2
        else:
            COLOR = 1
            LOCAL_LEADER = 0
            REMOTE_LEADER = 0
        INTRACOMM = BASECOMM.Split(COLOR, key=0)
        INTERCOMM = MPI.Intracomm.Create_intercomm(
            INTRACOMM,
            LOCAL_LEADER,
            BASECOMM,
            REMOTE_LEADER,
        )
        INTRACOMM.Free()
        cls.COMM = INTERCOMM
        cls.COMM.Set_errhandler(MPI.ERRORS_RETURN)

    @classmethod
    def tearDownClass(cls):
        cls.COMM.Free()


if __name__ == "__main__":
    unittest.main()



import sys
import threading
import warnings

import mpiunittest as unittest

from mpi4py import MPI


def allocate(n):
    return bytearray(n)


_basic = [
    None,
    True,
    False,
    -7,
    0,
    7,
    -(2**63) + 1,
    2**63 - 1,
    -2.17,
    0.0,
    +1.62,
    1 + 2j,
    2 - 3j,
    "mpi4py",
]
messages = list(_basic)
messages += [
    list(_basic),
    tuple(_basic),
    set(_basic),
    frozenset(_basic),
    {f"k{k}": v for k, v in enumerate(_basic)},
]


class BaseTestP2PObj:
    #
    COMM = MPI.COMM_NULL

    def testSendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            self.COMM.send(smess, MPI.PROC_NULL)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertIsNone(rmess)
        if size == 1:
            return
        for smess in messages:
            if rank == 0:
                self.COMM.send(smess, rank + 1, 0)
                rmess = smess
            elif rank == size - 1:
                rmess = self.COMM.recv(None, rank - 1, 0)
            else:
                rmess = self.COMM.recv(None, rank - 1, 0)
                self.COMM.send(rmess, rank + 1, 0)
            self.assertEqual(rmess, smess)

    def testISendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        buf = None
        for smess in messages:
            req = self.COMM.isend(smess, MPI.PROC_NULL)
            self.assertTrue(req)
            req.Wait()
            self.assertFalse(req)
            rmess = self.COMM.recv(buf, MPI.PROC_NULL, 0)
            self.assertIsNone(rmess)
        for smess in messages:
            req = self.COMM.isend(smess, rank, 0)
            self.assertTrue(req)
            rmess = self.COMM.recv(buf, rank, 0)
            self.assertTrue(req)
            flag = False
            while not flag:
                flag = req.Test()
            self.assertTrue(flag)
            self.assertFalse(req)
            self.assertEqual(rmess, smess)
        for smess in messages:
            dst = (rank + 1) % size
            src = (rank - 1) % size
            req = self.COMM.isend(smess, dst, 0)
            self.assertTrue(req)
            rmess = self.COMM.recv(buf, src, 0)
            req.Wait()
            self.assertFalse(req)
            self.assertEqual(rmess, smess)

    def testIRecvAndSend(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        for smess in messages:
            req = comm.irecv(0, MPI.PROC_NULL)
            self.assertTrue(req)
            comm.send(smess, MPI.PROC_NULL)
            rmess = req.wait()
            self.assertFalse(req)
            self.assertIsNone(rmess)
        for smess in messages:
            buf = allocate(512)
            req = comm.irecv(buf, rank, 0)
            self.assertTrue(req)
            flag, rmess = req.test()
            self.assertTrue(req)
            self.assertFalse(flag)
            self.assertIsNone(rmess)
            comm.send(smess, rank, 0)
            self.assertTrue(req)
            flag, rmess = req.test()
            while not flag:
                flag, rmess = req.test()
            self.assertTrue(flag)
            self.assertFalse(req)
            self.assertEqual(rmess, smess)
        tmp = allocate(1024)
        for buf in (None, 1024, tmp):
            for smess in [*messages, messages]:
                dst = (rank + 1) % size
                src = (rank - 1) % size
                req = comm.irecv(buf, src, 0)
                self.assertTrue(req)
                comm.send(smess, dst, 0)
                rmess = req.wait()
                self.assertFalse(req)
                self.assertEqual(rmess, smess)
        for smess in messages:
            src = dst = rank
            rreq1 = comm.irecv(None, src, 1)
            rreq2 = comm.irecv(None, src, 2)
            rreq3 = comm.irecv(None, src, 3)
            rreqs = [rreq1, rreq2, rreq3]
            for i in range(len(rreqs)):
                self.assertTrue(rreqs[i])
                comm.send(smess, dst, i + 1)
                index, obj = MPI.Request.waitany(rreqs)
                self.assertEqual(index, i)
                self.assertEqual(obj, smess)
                self.assertFalse(rreqs[index])
            index, obj = MPI.Request.waitany(rreqs)
            self.assertEqual(index, MPI.UNDEFINED)
            self.assertIsNone(obj)
        for smess in messages:
            src = dst = rank
            rreq1 = comm.irecv(None, src, 1)
            rreq2 = comm.irecv(None, src, 2)
            rreq3 = comm.irecv(None, src, 3)
            rreqs = [rreq1, rreq2, rreq3]
            index, flag, obj = MPI.Request.testany(rreqs)
            self.assertEqual(index, MPI.UNDEFINED)
            self.assertFalse(flag)
            self.assertIsNone(obj)
            for i in range(len(rreqs)):
                self.assertTrue(rreqs[i])
                comm.send(smess, dst, i + 1)
                index, flag, obj = MPI.Request.testany(rreqs)
                while not flag:
                    index, flag, obj = MPI.Request.testany(rreqs)
                self.assertEqual(index, i)
                self.assertTrue(flag)
                self.assertEqual(obj, smess)
                self.assertFalse(rreqs[i])
            index, flag, obj = MPI.Request.testany(rreqs)
            self.assertEqual(index, MPI.UNDEFINED)
            self.assertTrue(flag)
            self.assertIsNone(obj)

    def testIRecvAndISend(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        tmp = allocate(512)
        for smess in messages:
            dst = (rank + 1) % size
            src = (rank - 1) % size
            rreq = comm.irecv(None, src, 0)
            self.assertTrue(rreq)
            sreq = comm.isend(smess, dst, 0)
            self.assertTrue(sreq)
            index1, mess1 = MPI.Request.waitany([sreq, rreq])
            self.assertIn(index1, (0, 1))
            if index1 == 0:
                self.assertFalse(sreq)
                self.assertTrue(rreq)
                self.assertIsNone(mess1)
            else:
                self.assertTrue(sreq)
                self.assertFalse(rreq)
                self.assertEqual(mess1, smess)
            index2, mess2 = MPI.Request.waitany([sreq, rreq])
            self.assertIn(index2, (0, 1))
            self.assertNotEqual(index2, index1)
            self.assertFalse(sreq)
            self.assertFalse(rreq)
            if index2 == 0:
                self.assertIsNone(mess2)
            else:
                self.assertEqual(mess2, smess)
        for smess in messages:
            dst = (rank + 1) % size
            src = (rank - 1) % size
            rreq = comm.irecv(None, src, 0)
            self.assertTrue(rreq)
            sreq = comm.isend(smess, dst, 0)
            self.assertTrue(sreq)
            index1, flag1, mess1 = MPI.Request.testany([sreq, rreq])
            while not flag1:
                index1, flag1, mess1 = MPI.Request.testany([sreq, rreq])
            self.assertIn(index1, (0, 1))
            if index1 == 0:
                self.assertFalse(sreq)
                self.assertTrue(rreq)
                self.assertIsNone(mess1)
            else:
                self.assertTrue(sreq)
                self.assertFalse(rreq)
                self.assertEqual(mess1, smess)
            index2, flag2, mess2 = MPI.Request.testany([sreq, rreq])
            while not flag2:
                index2, flag2, mess2 = MPI.Request.testany([sreq, rreq])
            self.assertIn(index2, (0, 1))
            self.assertNotEqual(index2, index1)
            self.assertFalse(sreq)
            self.assertFalse(rreq)
            if index2 == 0:
                self.assertIsNone(mess2)
            else:
                self.assertEqual(mess2, smess)
        for buf in (None, 512, tmp):
            for smess in messages:
                dst = (rank + 1) % size
                src = (rank - 1) % size
                rreq = comm.irecv(buf, src, 0)
                self.assertTrue(rreq)
                sreq = comm.isend(smess, dst, 0)
                self.assertTrue(sreq)
                dummy, rmess = MPI.Request.waitall([sreq, rreq], [])
                self.assertFalse(sreq)
                self.assertFalse(rreq)
                self.assertIsNone(dummy)
                self.assertEqual(rmess, smess)
        for buf in (None, 512, tmp):
            for smess in messages:
                src = dst = rank
                rreq = comm.irecv(buf, src, 1)
                flag, msg = MPI.Request.testall([rreq])
                self.assertFalse(flag)
                self.assertIsNone(msg)
                sreq = comm.isend(smess, dst, 1)
                while True:
                    flag, msg = MPI.Request.testall([sreq, rreq], [])
                    if not flag:
                        self.assertIsNone(msg)
                        continue
                    (dummy, rmess) = msg
                    self.assertIsNone(dummy)
                    self.assertEqual(rmess, smess)
                    break

    def testManyISendAndRecv(self):
        rank = self.COMM.Get_rank()
        for smess in messages:
            reqs = []
            for _k in range(6):
                r = self.COMM.isend(smess, rank, 0)
                reqs.append(r)
            flag = MPI.Request.Testall(reqs)
            if not flag:
                index, flag = MPI.Request.Testany(reqs)
                if not flag:
                    self.assertEqual(index, MPI.UNDEFINED)
                indices = MPI.Request.Testsome(reqs)
                self.assertIsInstance(indices, list)
            for _k in range(3):
                rmess = self.COMM.recv(None, rank, 0)
                self.assertEqual(rmess, smess)
            flag = MPI.Request.Testall(reqs)
            if not flag:
                index, flag = MPI.Request.Testany(reqs)
                self.assertEqual(index, 0)
                self.assertTrue(flag)
                indices = MPI.Request.Testsome(reqs)
                if indices is None:
                    count = MPI.UNDEFINED
                    indices = []
                else:
                    count = len(indices)
                    indices = sorted(indices)
                self.assertGreaterEqual(count, 2)
                self.assertEqual(indices[:2], [1, 2])
            for _k in range(3):
                rmess = self.COMM.recv(None, rank, 0)
                self.assertEqual(rmess, smess)
            flag = MPI.Request.Testall(reqs)
            self.assertTrue(flag)
        for smess in messages:
            reqs = []
            for _k in range(6):
                r = self.COMM.isend(smess, rank, 0)
                reqs.append(r)
            for _k in range(3):
                rmess = self.COMM.recv(None, rank, 0)
                self.assertEqual(rmess, smess)
            index = MPI.Request.Waitany(reqs)
            self.assertEqual(index, 0)
            indices1 = MPI.Request.Waitsome(reqs)
            if indices1 is None:
                count1 = MPI.UNDEFINED
                indices1 = []
            else:
                count1 = len(indices1)
            for _k in range(3):
                rmess = self.COMM.recv(None, rank, 0)
                self.assertEqual(rmess, smess)
            indices2 = MPI.Request.Waitsome(reqs)
            if indices2 is None:
                count2 = MPI.UNDEFINED
                indices2 = []
            else:
                count2 = len(indices2)
            if count1 == MPI.UNDEFINED:
                count1 = 0
            if count2 == MPI.UNDEFINED:
                count2 = 0
            self.assertEqual(6, 1 + count1 + count2)
            indices = [0, *list(indices1), *list(indices2)]
            indices.sort()
            self.assertEqual(indices, list(range(6)))
            MPI.Request.Waitall(reqs)

    def testManyISSendAndRecv(self):
        rank = self.COMM.Get_rank()
        for smess in messages:
            reqs = []
            for _k in range(6):
                r = self.COMM.issend(smess, rank, 0)
                reqs.append(r)
            flag = MPI.Request.Testall(reqs)
            self.assertFalse(flag)
            for _k in range(3):
                rmess = self.COMM.recv(None, rank, 0)
                self.assertEqual(rmess, smess)
            flag = MPI.Request.Testall(reqs)
            self.assertFalse(flag)
            index, flag = MPI.Request.Testany(reqs)
            if flag:
                self.assertEqual(index, 0)
            indices = MPI.Request.Testsome(reqs)
            if indices is not None:
                target = [0, 1, 2]
                if flag:
                    del target[0]
                for index in indices:
                    self.assertIn(index, target)
            for _k in range(3):
                flag = MPI.Request.Testall(reqs)
                self.assertFalse(flag)
                rmess = self.COMM.recv(None, rank, 0)
                self.assertEqual(rmess, smess)
            flag = False
            for _k in range(10):
                flag |= MPI.Request.Testall(reqs)
                if flag:
                    break
            if unittest.is_mpi("impi") and sys.platform == "win32":
                flag |= MPI.Request.Waitall(reqs)
            self.assertTrue(flag)

    def testSSendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            self.COMM.ssend(smess, MPI.PROC_NULL)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertIsNone(rmess)
        if size == 1:
            return
        for smess in messages:
            if rank == 0:
                self.COMM.ssend(smess, rank + 1, 0)
                rmess = smess
            elif rank == size - 1:
                rmess = self.COMM.recv(None, rank - 1, 0)
            else:
                rmess = self.COMM.recv(None, rank - 1, 0)
                self.COMM.ssend(rmess, rank + 1, 0)
            self.assertEqual(rmess, smess)

    def testISSendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            req = self.COMM.issend(smess, MPI.PROC_NULL)
            self.assertTrue(req)
            req.Wait()
            self.assertFalse(req)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertIsNone(rmess)
        for smess in messages:
            req = self.COMM.issend(smess, rank, 0)
            self.assertTrue(req)
            flag = req.Test()
            self.assertFalse(flag)
            self.assertTrue(req)
            rmess = self.COMM.recv(None, rank, 0)
            self.assertTrue(req)
            flag = False
            while not flag:
                flag = req.Test()
            self.assertTrue(flag)
            self.assertFalse(req)
            self.assertEqual(rmess, smess)
        for smess in messages:
            dst = (rank + 1) % size
            src = (rank - 1) % size
            req = self.COMM.issend(smess, dst, 0)
            self.assertTrue(req)
            rmess = self.COMM.recv(None, src, 0)
            req.Wait()
            self.assertFalse(req)
            self.assertEqual(rmess, smess)

    def testCancel(self):
        rank = self.COMM.Get_rank()
        status = MPI.Status()
        for smess in messages:
            req = self.COMM.issend(smess, rank)
            self.assertTrue(req)
            req.cancel()
            flag = req.get_status(status)
            cancelled = status.Is_cancelled()
            self.assertTrue(req)
            if cancelled:
                self.assertTrue(flag)
                req.Free()
                self.assertFalse(req)
            else:
                self.assertFalse(flag)
                rmess = self.COMM.recv(None, rank, 0)
                flag = req.get_status()
                while not flag:
                    flag = req.get_status()
                self.assertTrue(flag)
                self.assertTrue(req)
                flag, _ = req.test()
                self.assertTrue(flag)
                self.assertFalse(req)
                self.assertEqual(rmess, smess)

    def testIRecvAndBSend(self):
        comm = self.COMM
        rank = comm.Get_rank()
        buf = MPI.Alloc_mem((1 << 16) + MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        try:
            for smess in messages:
                src = dst = rank
                req1 = comm.irecv(None, src, 1)
                req2 = comm.irecv(None, src, 2)
                req3 = comm.irecv(None, src, 3)
                comm.bsend(smess, dst, 3)
                comm.bsend(smess, dst, 2)
                comm.bsend(smess, dst, 1)
                self.assertEqual(smess, req3.wait())
                self.assertEqual(smess, req2.wait())
                self.assertEqual(smess, req1.wait())
                comm.bsend(smess, MPI.PROC_NULL, 3)
        finally:
            MPI.Detach_buffer()
            MPI.Free_mem(buf)

    def testIRecvAndIBSend(self):
        comm = self.COMM
        rank = comm.Get_rank()
        buf = MPI.Alloc_mem((1 << 16) + MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        try:
            for smess in messages:
                src = dst = rank
                req1 = comm.irecv(None, src, 1)
                req2 = comm.irecv(None, src, 2)
                req3 = comm.irecv(None, src, 3)
                req4 = comm.ibsend(smess, dst, 3)
                req5 = comm.ibsend(smess, dst, 2)
                req6 = comm.ibsend(smess, dst, 1)
                MPI.Request.waitall([req4, req5, req6])
                self.assertEqual(smess, req3.wait())
                self.assertEqual(smess, req2.wait())
                self.assertEqual(smess, req1.wait())
                comm.ibsend(smess, MPI.PROC_NULL, 3).wait()
        finally:
            MPI.Detach_buffer()
            MPI.Free_mem(buf)

    def testIRecvAndSSend(self):
        comm = self.COMM
        rank = comm.Get_rank()
        for smess in messages:
            src = dst = rank
            req1 = comm.irecv(None, src, 1)
            req2 = comm.irecv(None, src, 2)
            req3 = comm.irecv(None, src, 3)
            comm.ssend(smess, dst, 3)
            comm.ssend(smess, dst, 2)
            comm.ssend(smess, dst, 1)
            self.assertEqual(smess, req3.wait())
            self.assertEqual(smess, req2.wait())
            self.assertEqual(smess, req1.wait())
            comm.ssend(smess, MPI.PROC_NULL, 3)

    def testIRecvAndISSend(self):
        comm = self.COMM
        rank = comm.Get_rank()
        for smess in messages:
            src = dst = rank
            req1 = comm.irecv(None, src, 1)
            req2 = comm.irecv(None, src, 2)
            req3 = comm.irecv(None, src, 3)
            req4 = comm.issend(smess, dst, 3)
            req5 = comm.issend(smess, dst, 2)
            req6 = comm.issend(smess, dst, 1)
            MPI.Request.waitall([req4, req5, req6])
            self.assertEqual(smess, req3.wait())
            self.assertEqual(smess, req2.wait())
            self.assertEqual(smess, req1.wait())
            comm.issend(smess, MPI.PROC_NULL, 3).wait()

    def testSendrecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            dest = (rank + 1) % size
            source = (rank - 1) % size
            rmess = self.COMM.sendrecv(smess, dest, 0, None, source, 0)
            continue
            self.assertEqual(rmess, smess)
            rmess = self.COMM.sendrecv(None, dest, 0, None, source, 0)
            self.assertIsNone(rmess)
            rmess = self.COMM.sendrecv(
                smess, MPI.PROC_NULL, 0, None, MPI.PROC_NULL, 0
            )
            self.assertIsNone(rmess)

    def testMixed(self):
        comm = self.COMM
        rank = comm.Get_rank()
        #
        sreq = comm.Isend([None, 0, "B"], rank)
        obj = comm.recv(None, rank)
        sreq.Wait()
        self.assertIsNone(obj)
        for smess in messages:
            buf = MPI.pickle.dumps(smess)
            sreq = comm.Isend([buf, "B"], rank)
            rmess = comm.recv(None, rank)
            sreq.Wait()
            self.assertEqual(rmess, smess)
        #
        sreq = comm.Isend([None, 0, "B"], rank)
        rreq = comm.irecv(None, rank)
        sreq.Wait()
        obj = rreq.wait()
        self.assertIsNone(obj)
        for smess in messages:
            buf = MPI.pickle.dumps(smess)
            sreq = comm.Isend([buf, "B"], rank)
            rreq = comm.irecv(None, rank)
            sreq.Wait()
            rmess = rreq.wait()
            self.assertEqual(rmess, smess)

    def testPingPong01(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            self.COMM.send(smess, MPI.PROC_NULL)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertIsNone(rmess)
        if size == 1:
            return
        smess = None
        if rank == 0:
            self.COMM.send(smess, rank + 1, 0)
            rmess = self.COMM.recv(None, rank + 1, 0)
        elif rank == 1:
            rmess = self.COMM.recv(None, rank - 1, 0)
            self.COMM.send(smess, rank - 1, 0)
        else:
            rmess = smess
        self.assertEqual(rmess, smess)
        for smess in messages:
            if rank == 0:
                self.COMM.send(smess, rank + 1, 0)
                rmess = self.COMM.recv(None, rank + 1, 0)
            elif rank == 1:
                rmess = self.COMM.recv(None, rank - 1, 0)
                self.COMM.send(smess, rank - 1, 0)
            else:
                rmess = smess
            self.assertEqual(rmess, smess)

    @unittest.skipMPI("MPICH1")
    def testProbe(self):
        comm = self.COMM.Dup()
        try:
            status = MPI.Status()
            flag = comm.iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertFalse(flag)
            for smess in messages:
                request = comm.issend(smess, comm.rank, 123)
                self.assertTrue(request)
                while not comm.iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status):
                    pass
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                comm.probe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertTrue(request)
                flag, obj = request.test()
                self.assertTrue(request)
                self.assertFalse(flag)
                self.assertIsNone(obj)
                obj = comm.recv(None, comm.rank, 123)
                self.assertEqual(obj, smess)
                self.assertTrue(request)
                obj = request.wait()
                self.assertFalse(request)
                self.assertIsNone(obj)
        finally:
            comm.Free()

    def testWaitSomeRecv(self):
        comm = self.COMM.Dup()
        rank = comm.Get_rank()
        reqs = [comm.irecv(source=rank, tag=i) for i in range(6)]
        for indexlist in ([5], [3, 1, 2], [0, 4]):
            for i in indexlist:
                comm.ssend("abc", dest=rank, tag=i)
            statuses = []
            idxs, objs = MPI.Request.waitsome(reqs, statuses)
            self.assertEqual(sorted(idxs), sorted(indexlist))
            self.assertEqual(objs, ["abc"] * len(idxs))
            self.assertFalse(any(reqs[i] for i in idxs))
            self.assertEqual(len(statuses), len(idxs))
            self.assertTrue(all(s.source == rank for s in statuses))
            self.assertTrue(all(s.tag in indexlist for s in statuses))
            self.assertTrue(all(s.error == MPI.SUCCESS for s in statuses))
        idxs, objs = MPI.Request.waitsome(reqs)
        self.assertIsNone(idxs)
        self.assertIsNone(objs)
        self.assertFalse(any(reqs))
        comm.Free()

    def testTestSomeRecv(self):
        comm = self.COMM.Dup()
        rank = comm.Get_rank()
        reqs = [comm.irecv(source=rank, tag=i) for i in range(6)]
        statuses = []
        idxs, objs = MPI.Request.testsome(reqs, statuses)
        self.assertEqual(idxs, [])
        self.assertEqual(objs, [])
        self.assertTrue(all(reqs))
        self.assertEqual(statuses, [])
        for indexlist in ([5], [], [3, 1, 2], [], [0, 4]):
            for i in indexlist:
                comm.ssend("abc", dest=rank, tag=i)
            statuses = []
            idxs, objs = MPI.Request.testsome(reqs, statuses)
            self.assertEqual(sorted(idxs), sorted(indexlist))
            self.assertEqual(objs, ["abc"] * len(idxs))
            self.assertFalse(any(reqs[i] for i in idxs))
            self.assertEqual(len(statuses), len(idxs))
            self.assertTrue(all(s.source == rank for s in statuses))
            self.assertTrue(all(s.tag in indexlist for s in statuses))
            self.assertTrue(all(s.error == MPI.SUCCESS for s in statuses))
        idxs, objs = MPI.Request.testsome(reqs)
        self.assertIsNone(idxs)
        self.assertIsNone(objs)
        self.assertFalse(any(reqs))
        comm.Free()

    def testWaitSomeSend(self):
        comm = self.COMM.Dup()
        rank = comm.Get_rank()
        reqs = [comm.issend("abc", dest=rank, tag=i) for i in range(6)]
        for indexlist in ([5], [3, 1, 2], [0, 4]):
            for i in indexlist:
                msg = comm.recv(source=rank, tag=i)
                self.assertEqual(msg, "abc")
            idxs, objs = MPI.Request.waitsome(reqs)
            while sorted(idxs) != sorted(indexlist):
                i, o = MPI.Request.waitsome(reqs)
                idxs.extend(i)
                objs.extend(o)
            self.assertEqual(sorted(idxs), sorted(indexlist))
            self.assertEqual(objs, [None] * len(idxs))
            self.assertFalse(any(reqs[i] for i in idxs))
        idxs, objs = MPI.Request.waitsome(reqs)
        self.assertIsNone(idxs)
        self.assertIsNone(objs)
        self.assertFalse(any(reqs))
        comm.Free()

    def testTestSomeSend(self):
        comm = self.COMM.Dup()
        rank = comm.Get_rank()
        reqs = [comm.issend("abc", dest=rank, tag=i) for i in range(6)]
        idxs, objs = MPI.Request.testsome(reqs)
        self.assertEqual(idxs, [])
        self.assertEqual(objs, [])
        self.assertTrue(all(reqs))
        for indexlist in ([5], [], [3, 1, 2], [], [0, 4]):
            for i in indexlist:
                msg = comm.recv(source=rank, tag=i)
                self.assertEqual(msg, "abc")
            idxs, objs = MPI.Request.testsome(reqs)
            while sorted(idxs) != sorted(indexlist):
                i, o = MPI.Request.testsome(reqs)
                idxs.extend(i)
                objs.extend(o)
            self.assertEqual(sorted(idxs), sorted(indexlist))
            self.assertEqual(objs, [None] * len(idxs))
            self.assertFalse(any(reqs[i] for i in idxs))
        idxs, objs = MPI.Request.testsome(reqs)
        self.assertIsNone(idxs)
        self.assertIsNone(objs)
        self.assertFalse(any(reqs))
        comm.Free()

    def testRecvObjArg(self):
        comm = self.COMM
        rank = comm.Get_rank()
        req1 = comm.isend("42", rank)
        req2 = comm.isend([42], rank)
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            with self.assertRaises(UserWarning):
                comm.recv(bytearray(0), MPI.PROC_NULL)
            warnings.simplefilter("ignore")
            obj = comm.recv(128, rank)
            self.assertEqual(obj, "42")
            req1.wait()
            obj = comm.recv(bytearray(128), rank)
            self.assertEqual(obj, [42])
            req2.wait()

    def testCommLock(self):
        comm = self.COMM.Dup()
        table = MPI._comm_lock_table(comm)
        self.assertIsInstance(table, dict)
        self.assertNotIn("bcast", table)
        comm.bcast(None, root=0)
        self.assertIn("bcast", table)
        lock = table["bcast"]
        lock_type = type(threading.Lock())
        self.assertIsInstance(lock, lock_type)
        comm.Free()


class TestP2PObjSelf(BaseTestP2PObj, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestP2PObjWorld(BaseTestP2PObj, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestP2PObjSelfDup(TestP2PObjSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


@unittest.skipMPI("openmpi(<1.4.0)", MPI.Query_thread() > MPI.THREAD_SINGLE)
class TestP2PObjWorldDup(TestP2PObjWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


if __name__ == "__main__":
    unittest.main()



# Opening modes
# -------------

MODE_RDONLY = MPI_MODE_RDONLY
#: Read only

MODE_WRONLY = MPI_MODE_WRONLY
#: Write only

MODE_RDWR  = MPI_MODE_RDWR
#: Reading and writing

MODE_CREATE = MPI_MODE_CREATE
#: Create the file if it does not exist

MODE_EXCL = MPI_MODE_EXCL
#: Error if creating file that already exists

MODE_DELETE_ON_CLOSE = MPI_MODE_DELETE_ON_CLOSE
#: Delete file on close

MODE_UNIQUE_OPEN = MPI_MODE_UNIQUE_OPEN
#: File will not be concurrently opened elsewhere

MODE_SEQUENTIAL = MPI_MODE_SEQUENTIAL
#: File will only be accessed sequentially

MODE_APPEND = MPI_MODE_APPEND
#: Set initial position of all file pointers to end of file


# Positioning
# -----------

SEEK_SET = MPI_SEEK_SET
#: File pointer is set to offset

SEEK_CUR = MPI_SEEK_CUR
#: File pointer is set to the current position plus offset

SEEK_END = MPI_SEEK_END
#: File pointer is set to the end plus offset

DISPLACEMENT_CURRENT = MPI_DISPLACEMENT_CURRENT
#: Special displacement value for files opened in sequential mode

DISP_CUR = MPI_DISPLACEMENT_CURRENT
#: Convenience alias for `DISPLACEMENT_CURRENT`


cdef class File:
    """
    File I/O context.
    """

    def __cinit__(self, File file: File | None = None):
        cinit(self, file)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, File): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_default(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> File:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_File> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Close` if not null.
        """
        safefree(self)

    # File Manipulation
    # -----------------

    @classmethod
    def Open(
        cls,
        Intracomm comm: Intracomm,
        filename: PathLike[AnyStr] | str | bytes,
        int amode: int = MODE_RDONLY,
        Info info: Info = INFO_NULL,
    ) -> Self:
        """
        Open a file.
        """
        cdef char *cfilename = NULL
        filename = asmpifspath(filename, &cfilename)
        cdef File file = <File>New(cls)
        with nogil: CHKERR( MPI_File_open(
            comm.ob_mpi, cfilename, amode, info.ob_mpi, &file.ob_mpi) )
        options_set_errhandler(file.ob_mpi)
        return file

    def Close(self) -> None:
        """
        Close a file.
        """
        cdef MPI_File save = self.ob_mpi
        with nogil: CHKERR( MPI_File_close(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save

    @classmethod
    def Delete(
        cls,
        filename: PathLike[AnyStr] | str | bytes,
        Info info: Info = INFO_NULL,
    ) -> None:
        """
        Delete a file.
        """
        cdef char *cfilename = NULL
        filename = asmpifspath(filename, &cfilename)
        with nogil: CHKERR( MPI_File_delete(cfilename, info.ob_mpi) )

    def Set_size(self, Offset size: int) -> None:
        """
        Set the file size.
        """
        with nogil: CHKERR( MPI_File_set_size(self.ob_mpi, size) )

    def Preallocate(self, Offset size: int) -> None:
        """
        Preallocate storage space for a file.
        """
        with nogil: CHKERR( MPI_File_preallocate(self.ob_mpi, size) )

    def Get_size(self) -> int:
        """
        Return the file size.
        """
        cdef MPI_Offset size = 0
        with nogil: CHKERR( MPI_File_get_size(self.ob_mpi, &size) )
        return size

    property size:
        """Size (in bytes)."""
        def __get__(self) -> int:
            return self.Get_size()

    def Get_amode(self) -> int:
        """
        Return the file access mode.
        """
        cdef int amode = 0
        with nogil: CHKERR( MPI_File_get_amode(self.ob_mpi, &amode) )
        return amode

    property amode:
        """Access mode."""
        def __get__(self) -> int:
            return self.Get_amode()

    # File Group
    # ----------

    def Get_group(self) -> Group:
        """
        Access the group of processes that opened the file.
        """
        cdef Group group = <Group>New(Group)
        with nogil: CHKERR( MPI_File_get_group(self.ob_mpi, &group.ob_mpi) )
        return group

    property group:
        """Group."""
        def __get__(self) -> Group:
            return self.Get_group()

    property group_size:
        """Group size."""
        def __get__(self) -> int:
            cdef MPI_Group group = MPI_GROUP_NULL
            cdef int group_size  = -1
            CHKERR( MPI_File_get_group(self.ob_mpi, &group) )
            try:
                CHKERR( MPI_Group_size(group, &group_size) )
            finally:
                CHKERR( MPI_Group_free(&group) )
            return group_size

    property group_rank:
        """Group rank."""
        def __get__(self) -> int:
            cdef MPI_Group group = MPI_GROUP_NULL
            cdef int group_rank  = MPI_PROC_NULL
            CHKERR( MPI_File_get_group(self.ob_mpi, &group) )
            try:
                CHKERR( MPI_Group_rank(group, &group_rank) )
            finally:
                CHKERR( MPI_Group_free(&group) )
            return group_rank

    # File Info
    # ---------

    def Set_info(self, Info info: Info) -> None:
        """
        Set new values for the hints associated with a file.
        """
        with nogil: CHKERR( MPI_File_set_info(self.ob_mpi, info.ob_mpi) )

    def Get_info(self) -> Info:
        """
        Return the current hints for a file.
        """
        cdef Info info = <Info>New(Info)
        with nogil: CHKERR( MPI_File_get_info(self.ob_mpi, &info.ob_mpi) )
        return info

    property info:
        """Info hints."""
        def __get__(self) -> Info:
            return self.Get_info()

        def __set__(self, value: Info):
            self.Set_info(value)

    # File Views
    # ----------

    def Set_view(
        self,
        Offset disp: int = 0,
        Datatype etype: Datatype = BYTE,
        Datatype filetype: Datatype | None = None,
        datarep: str = "native",
        Info info: Info = INFO_NULL,
    ) -> None:
        """
        Set the file view.
        """
        cdef char *cdatarep = b"native"
        if datarep is not None: datarep = asmpistr(datarep, &cdatarep)
        cdef MPI_Datatype cetype = etype.ob_mpi
        cdef MPI_Datatype cftype = cetype
        if filetype is not None: cftype = filetype.ob_mpi
        with nogil: CHKERR( MPI_File_set_view(
            self.ob_mpi, disp, cetype, cftype, cdatarep, info.ob_mpi) )

    def Get_view(self) -> tuple[int, Datatype, Datatype, str]:
        """
        Return the file view.
        """
        cdef MPI_Offset disp = 0
        cdef MPI_Datatype cetype = MPI_DATATYPE_NULL
        cdef MPI_Datatype cftype = MPI_DATATYPE_NULL
        cdef char cdatarep[MPI_MAX_DATAREP_STRING+1]
        cdatarep[0] = 0  # just in case
        with nogil: CHKERR( MPI_File_get_view(
            self.ob_mpi, &disp, &cetype, &cftype, cdatarep) )
        cdatarep[MPI_MAX_DATAREP_STRING] = 0  # just in case
        cdef Datatype etype = ref_Datatype(cetype)
        cdef Datatype ftype = ref_Datatype(cftype)
        cdef object datarep = mpistr(cdatarep)
        return (disp, etype, ftype, datarep)

    # Data Access
    # -----------

    # Data Access with Explicit Offsets
    # ---------------------------------

    def Read_at(
        self,
        Offset offset: int,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Read using explicit offset.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_at_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, statusp) )

    def Read_at_all(
        self,
        Offset offset: int,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Collective read using explicit offset.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_at_all_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, statusp) )

    def Write_at(
        self,
        Offset offset: int,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Write using explicit offset.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_at_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, statusp) )

    def Write_at_all(
        self,
        Offset offset: int,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Collective write using explicit offset.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_at_all_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, statusp) )

    def Iread_at(
        self,
        Offset offset: int,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking read using explicit offset.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iread_at_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Iread_at_all(
        self,
        Offset offset: int,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking collective read using explicit offset.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iread_at_all_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Iwrite_at(
        self,
        Offset offset: int,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking write using explicit offset.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iwrite_at_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Iwrite_at_all(
        self,
        Offset offset: int,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking collective write using explicit offset.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iwrite_at_all_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    # Data Access with Individual File Pointers
    # -----------------------------------------

    def Read(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Read using individual file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Read_all(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Collective read using individual file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_all_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Write(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Write using individual file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Write_all(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Collective write using individual file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_all_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Iread(
        self,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking read using individual file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iread_c(
            self.ob_mpi, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Iread_all(
        self,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking collective read using individual file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iread_all_c(
            self.ob_mpi, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Iwrite(
        self,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking write using individual file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iwrite_c(
            self.ob_mpi, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Iwrite_all(
        self,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking collective write using individual file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iwrite_all_c(
            self.ob_mpi, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Seek(self, Offset offset: int, int whence: int = SEEK_SET) -> None:
        """
        Update the individual file pointer.
        """
        with nogil: CHKERR( MPI_File_seek(self.ob_mpi, offset, whence) )

    def Get_position(self) -> int:
        """
        Return the current position of the individual file pointer.

        .. note:: Position is measured in etype units
           relative to the current file view.
        """
        cdef MPI_Offset offset = 0
        with nogil: CHKERR( MPI_File_get_position(self.ob_mpi, &offset) )
        return offset

    def Get_byte_offset(self, Offset offset: int) -> int:
        """
        Return the absolute byte position in the file.

        .. note:: Input *offset* is measured in etype units
           relative to the current file view.
        """
        cdef MPI_Offset disp = 0
        with nogil: CHKERR( MPI_File_get_byte_offset(
            self.ob_mpi, offset, &disp) )
        return disp

    # Data Access with Shared File Pointers
    # -------------------------------------

    def Read_shared(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Read using shared file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_shared_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Write_shared(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Write using shared file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_shared_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Iread_shared(
        self,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking read using shared file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iread_shared_c(
            self.ob_mpi, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Iwrite_shared(
        self,
        buf: BufSpec,
    ) -> Request:
        """
        Nonblocking write using shared file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_File_iwrite_shared_c(
            self.ob_mpi, m.buf, m.count, m.dtype, &request.ob_mpi) )
        request.ob_buf = m
        return request

    def Read_ordered(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Collective read using shared file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_ordered_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Write_ordered(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Collective write using shared file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_ordered_c(
            self.ob_mpi, m.buf, m.count, m.dtype, statusp) )

    def Seek_shared(
        self,
        Offset offset: int,
        int whence: int = SEEK_SET,
    ) -> None:
        """
        Update the shared file pointer.
        """
        with nogil: CHKERR( MPI_File_seek_shared(
            self.ob_mpi, offset, whence) )

    def Get_position_shared(self) -> int:
        """
        Return the current position of the shared file pointer.

        .. note:: Position is measured in etype units
           relative to the current view.
        """
        cdef MPI_Offset offset = 0
        with nogil: CHKERR( MPI_File_get_position_shared(
            self.ob_mpi, &offset) )
        return offset

    # Split Collective Data Access Routines
    # -------------------------------------

    # explicit offset

    def Read_at_all_begin(
        self,
        Offset offset: int,
        buf: BufSpec,
    ) -> None:
        """
        Start a split collective read using explicit offset.
        """
        cdef _p_msg_io m = message_io_read(buf)
        with nogil: CHKERR( MPI_File_read_at_all_begin_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype) )

    def Read_at_all_end(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Complete a split collective read using explicit offset.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_at_all_end(
            self.ob_mpi, m.buf, statusp) )

    def Write_at_all_begin(
        self,
        Offset offset: int,
        buf: BufSpec,
    ) -> None:
        """
        Start a split collective write using explicit offset.
        """
        cdef _p_msg_io m = message_io_write(buf)
        with nogil: CHKERR( MPI_File_write_at_all_begin_c(
            self.ob_mpi, offset, m.buf, m.count, m.dtype) )

    def Write_at_all_end(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Complete a split collective write using explicit offset.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_at_all_end(
            self.ob_mpi, m.buf, statusp) )

    # individual file pointer

    def Read_all_begin(
        self,
        buf: BufSpec,
    ) -> None:
        """
        Start a split collective read using individual file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        with nogil: CHKERR( MPI_File_read_all_begin_c(
            self.ob_mpi, m.buf, m.count, m.dtype) )

    def Read_all_end(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Complete a split collective read using individual file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_all_end(
            self.ob_mpi, m.buf, statusp) )

    def Write_all_begin(
        self,
        buf: BufSpec,
    ) -> None:
        """
        Start a split collective write using individual file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        with nogil: CHKERR( MPI_File_write_all_begin_c(
            self.ob_mpi, m.buf, m.count, m.dtype) )

    def Write_all_end(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Complete a split collective write using individual file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_all_end(
            self.ob_mpi, m.buf, statusp) )

    # shared file pointer

    def Read_ordered_begin(
        self,
        buf: BufSpec,
    ) -> None:
        """
        Start a split collective read using shared file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        with nogil: CHKERR( MPI_File_read_ordered_begin_c(
            self.ob_mpi, m.buf, m.count, m.dtype) )

    def Read_ordered_end(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Complete a split collective read using shared file pointer.
        """
        cdef _p_msg_io m = message_io_read(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_read_ordered_end(
            self.ob_mpi, m.buf, statusp) )

    def Write_ordered_begin(
        self,
        buf: BufSpec,
    ) -> None:
        """
        Start a split collective write using shared file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        with nogil: CHKERR( MPI_File_write_ordered_begin_c(
            self.ob_mpi, m.buf, m.count, m.dtype) )

    def Write_ordered_end(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Complete a split collective write using shared file pointer.
        """
        cdef _p_msg_io m = message_io_write(buf)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_File_write_ordered_end(
            self.ob_mpi, m.buf, statusp) )

    # File Interoperability
    # ---------------------

    def Get_type_extent(self, Datatype datatype: Datatype) -> int:
        """
        Return the extent of datatype in the file.
        """
        cdef MPI_Count extent = 0
        with nogil: CHKERR( MPI_File_get_type_extent_c(
            self.ob_mpi, datatype.ob_mpi, &extent) )
        return extent

    # Consistency and Semantics
    # -------------------------

    def Set_atomicity(self, bint flag: bool) -> None:
        """
        Set the atomicity mode.
        """
        with nogil: CHKERR( MPI_File_set_atomicity(self.ob_mpi, flag) )

    def Get_atomicity(self) -> bool:
        """
        Return the atomicity mode.
        """
        cdef int flag = 0
        with nogil: CHKERR( MPI_File_get_atomicity(self.ob_mpi, &flag) )
        return <bint>flag

    property atomicity:
        """Atomicity mode."""
        def __get__(self) -> bool:
            return self.Get_atomicity()

        def __set__(self, value: bool):
            self.Set_atomicity(value)

    def Sync(self) -> None:
        """
        Causes all previous writes to be transferred to the storage device.
        """
        with nogil: CHKERR( MPI_File_sync(self.ob_mpi) )

    # Error Handling
    # --------------

    @classmethod
    def Create_errhandler(
        cls,
        errhandler_fn: Callable[[File, int], None],
    ) -> Errhandler:
        """
        Create a new error handler for files.
        """
        cdef Errhandler errhandler = <Errhandler>New(Errhandler)
        cdef MPI_File_errhandler_function *fn = NULL
        cdef int index = errhdl_new(errhandler_fn, &fn)
        try:
            CHKERR( MPI_File_create_errhandler(fn, &errhandler.ob_mpi) )
        except:                     # ~> uncovered  # noqa
            errhdl_del(&index, fn)  # ~> uncovered
            raise                   # ~> uncovered
        return errhandler

    def Get_errhandler(self) -> Errhandler:
        """
        Get the error handler for a file.
        """
        cdef Errhandler errhandler = <Errhandler>New(Errhandler)
        CHKERR( MPI_File_get_errhandler(self.ob_mpi, &errhandler.ob_mpi) )
        return errhandler

    def Set_errhandler(self, Errhandler errhandler: Errhandler) -> None:
        """
        Set the error handler for a file.
        """
        CHKERR( MPI_File_set_errhandler(self.ob_mpi, errhandler.ob_mpi) )

    def Call_errhandler(self, int errorcode: int) -> None:
        """
        Call the error handler installed on a file.
        """
        CHKERR( MPI_File_call_errhandler(self.ob_mpi, errorcode) )

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_File_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> File:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_File_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_File_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> File:
        """
        """
        return fromhandle(MPI_File_f2c(arg))


cdef File __FILE_NULL__ = def_File( MPI_FILE_NULL , "FILE_NULL" )


# Predefined file handles
# -----------------------

FILE_NULL = __FILE_NULL__  #: Null file handle


# User-defined data representations
# ---------------------------------

def Register_datarep(
    datarep: str,
    read_fn: Callable[[Buffer, Datatype, int, Buffer, int], None],
    write_fn: Callable[[Buffer, Datatype, int, Buffer, int], None],
    extent_fn: Callable[[Datatype], int],
) -> None:
    """
    Register user-defined data representations.
    """
    cdef char *cdatarep = NULL
    datarep = asmpistr(datarep, &cdatarep)
    cdef object state = _p_datarep(read_fn, write_fn, extent_fn)
    cdef MPI_Datarep_conversion_function_c *rd = MPI_CONVERSION_FN_NULL_C
    cdef MPI_Datarep_conversion_function_c *wr = MPI_CONVERSION_FN_NULL_C
    cdef MPI_Datarep_extent_function       *ex = datarep_extent_fn
    cdef void *xs = <void*> state
    if read_fn  is not None: rd = datarep_read_fn
    if write_fn is not None: wr = datarep_write_fn
    CHKERR ( MPI_Register_datarep_c(cdatarep, rd, wr, ex, xs) )
    with datarep_lock: datarep_registry[datarep] = state



import sys
from typing import Any, Callable, Iterable, Iterator, Mapping, Sequence

if sys.version_info >= (3, 10):
    from typing import ParamSpec, TypeAlias
else:
    from typing_extensions import ParamSpec, TypeAlias
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ..MPI import COMM_WORLD, Intracomm
from ..typing import T
from ._base import Executor, Future

_P = ParamSpec("_P")

class MPIPoolExecutor(Executor):
    Future: TypeAlias = Future
    def __init__(
        self,
        max_workers: int | None = None,
        initializer: Callable[..., object] | None = None,
        initargs: Iterable[Any] = (),
        *,
        python_exe: str = ...,
        python_args: Sequence[str] = ...,
        mpi_info: Mapping[str, str] | Iterable[tuple[str, str]] = ...,
        globals: Mapping[str, str] | Iterable[tuple[str, str]] = ...,  # noqa: A002
        main: bool = True,
        path: Sequence[str] = ...,
        wdir: str = ...,
        env: Mapping[str, str] | Iterable[tuple[str, str]] = ...,
        **kwargs: Any,
    ) -> None: ...
    @property
    def num_workers(self) -> int: ...
    def bootup(
        self,
        wait: bool = True,
    ) -> Self: ...
    if sys.version_info >= (3, 9):
        def submit(
            self,
            fn: Callable[_P, T],
            /,
            *args: _P.args,
            **kwargs: _P.kwargs,
        ) -> Future[T]: ...

    else:
        def submit(
            self,
            fn: Callable[_P, T],
            *args: _P.args,
            **kwargs: _P.kwargs,
        ) -> Future[T]: ...
    def map(
        self,
        fn: Callable[..., T],
        *iterables: Iterable[Any],
        timeout: float | None = None,
        chunksize: int = 1,
        buffersize: int | None = None,
        unordered: bool = False,
    ) -> Iterator[T]: ...
    def starmap(
        self,
        fn: Callable[..., T],
        iterable: Iterable[Any],
        timeout: float | None = None,
        chunksize: int = 1,
        buffersize: int | None = None,
        unordered: bool = False,
    ) -> Iterator[T]: ...
    def shutdown(
        self,
        wait: bool = True,
        *,
        cancel_futures: bool = False,
    ) -> None: ...

class MPICommExecutor:
    def __init__(
        self,
        comm: Intracomm | None = COMM_WORLD,
        root: int = 0,
        **kwargs: Any,
    ) -> None: ...
    def __enter__(self) -> Self | None: ...
    def __exit__(self, *args: object) -> bool | None: ...

class ThreadPoolExecutor(MPIPoolExecutor): ...
class ProcessPoolExecutor(MPIPoolExecutor): ...

def get_comm_workers() -> Intracomm: ...
